#include "Application.h"

namespace TR
{
	Application::Application() :
		mRoot(NULL),
		mRenderWindow(NULL),
		mGameEngine(NULL),
		mShutdown(false),
		mController(NULL),
		mMenu(NULL),
		mMenuMgr(NULL),
#ifndef NO_3DMOUSE
		mMouse3dConnexion(NULL),
#endif
		mMouse(NULL),
		mKeyboard(NULL)
	{
		
	}

	Application::~Application()
	{
		delete mGameEngine;
		delete mRoot;
		delete mMenu;

#ifndef NO_3DMOUSE
		delete mMouse3dConnexion;
#endif
		delete mMouse;
		delete mKeyboard;
		delete mController;
	}

	void Application::run()
	{
		mRoot = new Ogre::Root("plugins.cfg", "config.cfg", "ogre.log");
		
		// Show the config dialog.
		if(mRoot->showConfigDialog() == false) 
			return;
		
		// Initialise application and render window.
		mRenderWindow = mRoot->initialise(true, "Tamax & Ruby");
		mRoot->addFrameListener(this);
		Ogre::WindowEventUtilities::addWindowEventListener(mRenderWindow, this);

		// Load resources from resources.cfg.
		this->loadResources();

		// Create the scene.
		this->createScene();

		// Start rendering the scene.
 		mRoot->startRendering();
	}

	void Application::createScene()
	{
		// Create the controllers.
#ifndef NO_3DMOUSE
		mMouse3dConnexion = new Mouse3dConnexionController();
#endif
		mMouse = new MouseController();
		mKeyboard = new KeyboardController();

#ifndef NO_3DMOUSE
		mMouse3dConnexion->init(mRenderWindow);
#endif
		mMouse->init(mRenderWindow);
		mKeyboard->init(mRenderWindow);
		mController = new ControllerConfiguration();

		// Default control configuration.
		mController->setController1(mKeyboard); // Character control.
		mController->setController2(mMouse); // Camera control.
		mController->setController3(mKeyboard); // Menu control.
		
		// Create the menu.
		mMenuMgr = mRoot->createSceneManager(Ogre::ST_GENERIC, "MenuMgr");
		Ogre::Camera* cam = mMenuMgr->createCamera("menuCam");
		Ogre::Viewport* viewport = mRenderWindow->addViewport(cam);
		viewport->setBackgroundColour(Ogre::ColourValue(0,0,0));
		mMenu = new Menu(Ogre::ColourValue::White, Ogre::ColourValue(0.93,0.88,0.30));
		mMenu->initialize(mKeyboard->getKeyboard());
		mMenu->open();
	}

	bool Application::frameStarted(const Ogre::FrameEvent &evt)
	{
		// Controller capture.
		mController->prepareCapture();
		mController->capture();

		// Update the menu.
		mMenu->update();
		if(mShutdown == false) 
			mShutdown = mMenu->getOption(Menu::OP_QUIT) == Menu::QUIT_YES;

		// Update controllers.
		if(mMenu->getOption(Menu::OP_EVT) == Menu::EVT_DEVICE_CHANGED_YES)
		{
			if(mMenu->getOption(Menu::OP_DEVICE_CHARACTER) == Menu::DEVICE_KEYBOARD)
				mController->setController1(mKeyboard);
			else if(mMenu->getOption(Menu::OP_DEVICE_CHARACTER) == Menu::DEVICE_MOUSE)
				mController->setController1(mMouse);
#ifndef NO_3DMOUSE
			else if(mMenu->getOption(Menu::OP_DEVICE_CHARACTER) == Menu::DEVICE_3DMOUSE)
				mController->setController1(mMouse3dConnexion);
#endif	

			if(mMenu->getOption(Menu::OP_DEVICE_CAMERA) == Menu::DEVICE_KEYBOARD)
				mController->setController2(mKeyboard);
			else if(mMenu->getOption(Menu::OP_DEVICE_CAMERA) == Menu::DEVICE_MOUSE)
				mController->setController2(mMouse);
#ifndef NO_3DMOUSE
			else if(mMenu->getOption(Menu::OP_DEVICE_CAMERA) == Menu::DEVICE_3DMOUSE)
				mController->setController2(mMouse3dConnexion);
#endif
		}

		// Choose the game to create.
		if(mMenu->getOption(Menu::OP_NEW_GAME) == Menu::NEW_GAME_RUBY)
		{
			// Ruby game.
			if(mMenuMgr != NULL)
			{
				mRenderWindow->removeAllViewports();
				mRoot->destroySceneManager(mMenuMgr);
				mMenuMgr = NULL;
			}
			if(mGameEngine != NULL)
			{
				delete mGameEngine;
				mGameEngine = NULL;
			}
			mGameEngine = new RubyGameEngine();
			mGameEngine->createGame(mRenderWindow, mRoot);
			mGameEngine->setControllerConfiguration(mController);
		}
		else if(mMenu->getOption(Menu::OP_NEW_GAME) == Menu::NEW_GAME_TAMAX)
		{
			// Tamax game.

		}

		// If the game isn't paused.
		if(mMenu->isOpen() == false && mGameEngine != NULL)
		{
			// Update the game.
			mGameEngine->frameStarted(evt);
			
			// Update the rendering system.
			mGameEngine->getGameRenderer()->update(evt.timeSinceLastFrame);

			// Update the sound system.
			mGameEngine->getSoundManager()->FrameStarted(mGameEngine->getGameRenderer()->getEarNode(), evt.timeSinceLastFrame);
		}

		return !mShutdown;
	}

	bool Application::frameEnded(const Ogre::FrameEvent &evt)
	{
		// Update the game.
		if(mGameEngine != NULL)
			mGameEngine->frameEnded(evt);
		
		return !mShutdown;
	}

	void Application::windowResized(Ogre::RenderWindow* rw)
	{

	}
	
	void Application::windowClosed(Ogre::RenderWindow* rw)
	{
		mShutdown = true;
	}
	
	void Application::windowFocusChange(Ogre::RenderWindow* rw)
	{

	}

	void Application::loadResources()
	{
		// Load resource paths from config file
		Ogre::ConfigFile cf;
		cf.load("resources.cfg");

		// Go through all sections & settings in the file
		Ogre::ConfigFile::SectionIterator seci = cf.getSectionIterator();

		Ogre::String secName, typeName, archName;
		while (seci.hasMoreElements())
		{
			secName = seci.peekNextKey();
			Ogre::ConfigFile::SettingsMultiMap *settings = seci.getNext();
			Ogre::ConfigFile::SettingsMultiMap::iterator i;
			for (i = settings->begin(); i != settings->end(); ++i)
			{
				typeName = i->first;
				archName = i->second;
				Ogre::ResourceGroupManager::getSingleton().addResourceLocation(
					archName, typeName, secName);
			}
		}

		Ogre::ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
	}
}
