#include "Ruby.h"
#include "RubyBody.h"
#include "TerrainScroller.h"
#include "GameEngine.h"
#include "Properties.h"

namespace TR
{
	Ruby::Ruby(GameEngine* gameEngine):
		Character(gameEngine),
		mPreviousInputSpeed(Ogre::Vector3::ZERO),
		mCrashed(false),
		mIsNearTerrain(false),
		mAngleZ(0),
		mAngleX(0),
		mCrashSoundPlayed(false)
	{
		mBody = new RubyBody(this);
	}

	Ruby::~Ruby()
	{
	
	}

	void Ruby::updateState(const Ogre::Vector3& input,  double elapsedTime)
	{
		if(!mCrashed)
		{
			// Calculate desired speed.
			Ogre::Vector3 desiredInputSpeed;
			desiredInputSpeed = input * Properties::MAX_INPUT_SPEED;

			// Calculate input acceleration.
			Ogre::Vector3 inputAccel = desiredInputSpeed - mPreviousInputSpeed;
			if(inputAccel.squaredLength() > Properties::MAX_INPUT_ACCEL * Properties::MAX_INPUT_ACCEL)
			{
				inputAccel.normalise();
				inputAccel *= Properties::MAX_INPUT_ACCEL;
			}
			mPreviousInputSpeed += inputAccel * elapsedTime;

			// Calculate real acceleration.
			/*double portance = 9.81 / STABILITY_SPEED / STABILITY_SPEED;
			Ogre::Vector3 accel(inputAccel);
			double r = -9.81 + portance * mLinearSpeed.z * mLinearSpeed.z;
			if(r < 0)
				accel.y += r;
			accel.z = -inputAccel.y * 0.5 - 0.2;*/
	
			Ogre::Vector3 accel(inputAccel);
			Ogre::Vector3 gravity = Ogre::Vector3(0, -Properties::GRAVITY_FACTOR, 0);
			
			double tmp = mLinearSpeed.z/Properties::STABILITY_SPEED;
			if(tmp > 1)
				tmp = 1;

			Ogre::Vector3 portance = Ogre::Vector3(0, Properties::GRAVITY_FACTOR * tmp, 0);
			Ogre::Vector3 repulsion = Ogre::Vector3(0, 0, -(mLinearSpeed.z*mLinearSpeed.z / 5000)-(mLinearSpeed.y / 5));

			accel += (gravity + portance + repulsion);

			// Slows the up speed when the horizontale speed slows too much
			tmp = mLinearSpeed.z/20;
			if(tmp > 1)
				tmp = 1;
			
			if(accel.y > 0)
				accel.y *= tmp;

			// Resolve speed and position.
			mLinearSpeed += accel * elapsedTime;

			if(mLinearSpeed.y > 0 && desiredInputSpeed.y < 0)
				mLinearSpeed.y *= 0.9;

			if(mLinearSpeed.z < 0) 
				mLinearSpeed.z = 0;

				
			mPosition.y += mLinearSpeed.y * elapsedTime;
			// Y & Z are only used to move the terrain and the sky, not the character.
		}
		else
		{
			mLinearSpeed = Ogre::Vector3::ZERO;
		}

		// Calculate the corresponding rotation (graphical purpose only).
		mOrientation = Ogre::Quaternion(Ogre::Degree(-mAngleZ), Ogre::Vector3(0, 0, -1)) * mOrientation;
		mOrientation = Ogre::Quaternion(Ogre::Degree(-mAngleX), Ogre::Vector3(1, 0, 0)) * mOrientation;
		mAngleZ = (mLinearSpeed.x / Properties::MAX_INPUT_SPEED) * Properties::MAX_ROTATION_ANGLE;
		mAngleX = (mLinearSpeed.y / Properties::MAX_INPUT_SPEED) * Properties::MAX_ROTATION_ANGLE;
		mOrientation = Ogre::Quaternion(Ogre::Degree(mAngleX), Ogre::Vector3(1, 0, 0)) * mOrientation;
		mOrientation = Ogre::Quaternion(Ogre::Degree(mAngleZ), Ogre::Vector3(0, 0, -1)) * mOrientation;

		// Particle effects.
		mIsNearTerrain = mPosition.y < Properties::DUST_EFFECT_ALTI;
	}


	bool Ruby::checkCrash(TerrainScroller* terrain)
	{
		// Retrieves the front bottom point of Ruby
		Ogre::AxisAlignedBox aabb = mBody->getCollisionAABB();
		Ogre::Vector3 frontBottomPoint = aabb.getMinimum() + Ogre::Vector3(aabb.getHalfSize().x,0,0);

		// If Ruby is upside the terrain return false
		if(frontBottomPoint.y > terrain->getCenter().y + terrain->getMaximumY())
		{
			return false;
		}

		// Cheks all the meshes of the terrain
		Ogre::Node::ChildNodeIterator iterator = terrain->getSceneNode()->getChildIterator();
		while(iterator.hasMoreElements())
		{
			Ogre::SceneNode* subNode = static_cast<Ogre::SceneNode*>(iterator.getNext());
			Ogre::Entity* entity = static_cast<Ogre::Entity*>(subNode->getAttachedObject(0));

			// If Ruby intersects a part of the terrain
			if(aabb.intersects(entity->getWorldBoundingBox()))
			{
				// Throws a ray from Ruby to the terrain
				Ogre::Ray ray(frontBottomPoint, Ogre::Vector3::UNIT_Y);

				size_t vertexCount, indexCount;
				Ogre::Vector3* vertices;
				unsigned long* indices;

				// Retrieves the mesh informations
				getMeshInformation(entity->getMesh().getPointer(), vertexCount, vertices, indexCount, indices, subNode->getPosition(), subNode->getOrientation(), subNode->_getDerivedScale());

				// Checks all triangles of the mesh
				for(int i=0; i<static_cast<int>(indexCount); i+=3)
				{
					// Checks for a hit against this triangle
					std::pair<bool, Ogre::Real> hit = Ogre::Math::intersects(ray, vertices[indices[i]], vertices[indices[i+1]], vertices[indices[i+2]], true, true);

					// If the ray hits a triangle, Ruby is downside the terrain : return true
					if(hit.first)
					{
						delete[] vertices;
						delete[] indices;
						mCrashed = true;
						playCrashSound();
						return true;
					}
				}

				delete[] vertices;
				delete[] indices;
			}
		}

		return false;
	}

	void Ruby::playCrashSound()
	{
		if(mCrashSoundPlayed == false)
		{
			int sound = mGameEngine->getSoundManager()->CreateSound(Ogre::String("crash.ogg"), SOUND_TYPE_3D_SOUND);
			int soundChannel = 1;
			mGameEngine->getSoundManager()->PlaySound(sound, mBody->getSceneNode(), &soundChannel);
			mCrashSoundPlayed = true;
		}
	}

	// Get the mesh information for the given mesh.
	// Code found in Wiki: www.ogre3d.org/wiki/index.php/RetrieveVertexData
	void Ruby::getMeshInformation(const Ogre::Mesh* const mesh,
                        size_t &vertex_count,
                        Ogre::Vector3* &vertices,
                        size_t &index_count,
                        unsigned long* &indices,
                        const Ogre::Vector3 &position,
                        const Ogre::Quaternion &orient,
						const Ogre::Vector3 &scale)
	{
		bool added_shared = false;
		size_t current_offset = 0;
		size_t shared_offset = 0;
		size_t next_offset = 0;
		size_t index_offset = 0;


		vertex_count = index_count = 0;

		// Calculate how many vertices and indices we're going to need
		for ( unsigned short i = 0; i < mesh->getNumSubMeshes(); ++i)
		{
			Ogre::SubMesh* submesh = mesh->getSubMesh(i);
			// We only need to add the shared vertices once
			if(submesh->useSharedVertices)
			{
				if( !added_shared )
				{
					vertex_count += mesh->sharedVertexData->vertexCount;
					added_shared = true;
				}
			}
			else
			{
				vertex_count += submesh->vertexData->vertexCount;
			}
			// Add the indices
			index_count += submesh->indexData->indexCount;
		}


		// Allocate space for the vertices and indices
		vertices = new Ogre::Vector3[vertex_count];
		indices = new unsigned long[index_count];

		added_shared = false;

		// Run through the submeshes again, adding the data into the arrays
		for (unsigned short i = 0; i < mesh->getNumSubMeshes(); ++i)
		{
			Ogre::SubMesh* submesh = mesh->getSubMesh(i);

			Ogre::VertexData* vertex_data = submesh->useSharedVertices ? mesh->sharedVertexData : submesh->vertexData;

			if ((!submesh->useSharedVertices) || (submesh->useSharedVertices && !added_shared))
			{
				if(submesh->useSharedVertices)
				{
					added_shared = true;
					shared_offset = current_offset;
				}

				const Ogre::VertexElement* posElem =
					vertex_data->vertexDeclaration->findElementBySemantic(Ogre::VES_POSITION);

				Ogre::HardwareVertexBufferSharedPtr vbuf =
					vertex_data->vertexBufferBinding->getBuffer(posElem->getSource());

				unsigned char* vertex =
					static_cast<unsigned char*>(vbuf->lock(Ogre::HardwareBuffer::HBL_READ_ONLY));

				// There is _no_ baseVertexPointerToElement() which takes an Ogre::Real or a double
				//  as second argument. So make it float, to avoid trouble when Ogre::Real will
				//  be comiled/typedefed as double:
				//Ogre::Real* pReal;
				float* pReal;

				for( size_t j = 0; j < vertex_data->vertexCount; ++j, vertex += vbuf->getVertexSize())
				{
					posElem->baseVertexPointerToElement(vertex, &pReal);
					Ogre::Vector3 pt(pReal[0], pReal[1], pReal[2]);
					vertices[current_offset + j] = (orient * (pt * scale)) + position;
				}
	            
				vbuf->unlock();
				next_offset += vertex_data->vertexCount;
			}


			Ogre::IndexData* index_data = submesh->indexData;
			size_t numTris = index_data->indexCount / 3;
			Ogre::HardwareIndexBufferSharedPtr ibuf = index_data->indexBuffer;
	        
			bool use32bitindexes = (ibuf->getType() == Ogre::HardwareIndexBuffer::IT_32BIT);

			unsigned long* pLong = static_cast<unsigned long*>(ibuf->lock(Ogre::HardwareBuffer::HBL_READ_ONLY));
			unsigned short* pShort = reinterpret_cast<unsigned short*>(pLong);

			size_t offset = (submesh->useSharedVertices)? shared_offset : current_offset;

			if ( use32bitindexes )
			{
				for ( size_t k = 0; k < numTris*3; ++k)
				{
					indices[index_offset++] = pLong[k] + static_cast<unsigned long>(offset);
				}
			}
			else
			{
				for ( size_t k = 0; k < numTris*3; ++k)
				{
					indices[index_offset++] = static_cast<unsigned long>(pShort[k]) +
											  static_cast<unsigned long>(offset);
				}
			}

			ibuf->unlock();
			current_offset = next_offset;
		}
	}

}