#include "RubyGameEngine.h"
#include "RubyGameRenderer.h"
#include "RubyEnvironment.h"
#include "Properties.h"

namespace TR
{
	RubyGameEngine::RubyGameEngine() :
		mRuby(NULL),
		mSky(NULL),
		mTerrain(NULL),
		mHud(NULL),
		mScoreMgr(NULL),
		mTimer(0)
	{

	}

	RubyGameEngine::~RubyGameEngine()
	{
		delete mRuby;
		delete mSky;
		delete mTerrain;
		delete mHud;
		delete mScoreMgr;
	}

	void RubyGameEngine::createGame(Ogre::RenderWindow* renderWindow, Ogre::Root* root)
	{
		GameEngine::createGame(renderWindow, root);

		// Build the game.
		mRuby = new Ruby(this);
		mRuby->translate(Ogre::Vector3(0, 50, 0));
		mRuby->accelerate(Ogre::Vector3(0, 0, 80));

		mSky = new SkyScroller(Ogre::Vector2(0, 0), 0.0006);
		mTerrain = new TerrainScroller(Ogre::Vector2(49.9, 49.9), // Tile size
									   Ogre::Vector2(20, 20), // Tiling
									   Ogre::Vector3(0,0,0),  // Center
									   Ogre::Vector3(0, 0, 1), // Depth direction
									   Ogre::Vector3(1, 0, 0)); // Width direction

		// Build the rendering system.
		mGameRenderer = new RubyGameRenderer(renderWindow, root, Ogre::ST_EXTERIOR_FAR, "DefaultScene");
		mGameRenderer->addToRenderQueue(mRuby->getObjectBody());
		mGameRenderer->setCharacterBody(mRuby->getObjectBody());

		mEnvironment = new RubyEnvironment(mGameRenderer, this);

		std::vector<std::pair<Ogre::String,double>> terrainMeshes;
		terrainMeshes.push_back(std::pair<Ogre::String,double>("Terrain_Type1.mesh",0.3));
		terrainMeshes.push_back(std::pair<Ogre::String,double>("Terrain_Type2.mesh",0.1));
		terrainMeshes.push_back(std::pair<Ogre::String,double>("Terrain_Type3.mesh",0.3)); 
		terrainMeshes.push_back(std::pair<Ogre::String,double>("Terrain_Type4.mesh",0.3));
		mTerrain->initialize(mGameRenderer->getSceneManager(), "terrain", terrainMeshes);
		
		mSky->initialize(mGameRenderer->getSceneManager(), "sky");

		// Build the sound system.
		mSoundManager->Initialize();

		// Information overlays.
		mScoreMgr = new RubyScoreManager();	
		mHud = new RubyHud();
		mHud->show(true);

		int sound = mSoundManager->CreateLoopedSound(Ogre::String("wind.ogg"));
		int soundChannel = 1;
		mSoundManager->PlaySound(sound, mRuby->getObjectBody()->getSceneNode(), &soundChannel);
			
		// Debug
		mTerrain->showBoundingBox(false);
		mEnvironment->showBoundingBox(false);
		mRuby->getObjectBody()->showBoundingBox(false);
	}

	bool RubyGameEngine::frameStarted(const Ogre::FrameEvent& evt)
	{
		Ogre::Vector2 rubySpeed(mRuby->getLinearSpeed().x, mRuby->getLinearSpeed().z);
		mTimer += evt.timeSinceLastFrame;

		// If the game is finished, don't do anything.
		if(mRuby->checkCrash(mTerrain) || mTimer > Properties::GAME_LENGTH)
		{
			mHud->showGameOver(evt.timeSinceLastFrame);
			return true;
		}

		// Character control.
		if(mController != NULL)
		{
			AbstractController* c1 = mController->getController1();
			Ogre::Vector3 input(c1->getTranslationX(), 
								c1->getTranslationY(),
								0);

			mRuby->updateState(input, evt.timeSinceLastFrame);
			mEnvironment->update(input, evt.timeSinceLastFrame, rubySpeed, mRuby->getPosition().y);
			mEnvironment->resolveCollisions(mRuby);
		}

		// Terrain and sky update.
		mTerrain->setSpeed(rubySpeed);
		mSky->setSpeed(rubySpeed);
		mTerrain->update(evt.timeSinceLastFrame);
		mSky->update(evt.timeSinceLastFrame);

		// Score and hud update.
		mScoreMgr->update(mRuby->getLinearSpeed(), mRuby->getPosition().y, evt.timeSinceLastFrame);
		mHud->update(mRuby->getLinearSpeed(), mRuby->getPosition().y, evt.timeSinceLastFrame, mScoreMgr->getScore());

		return true;
	}

	bool RubyGameEngine::frameEnded(const Ogre::FrameEvent& evt)
	{
		return true;
	}
}
