#include "RubyGameRenderer.h"
#include "Properties.h"

namespace TR
{
	RubyGameRenderer::RubyGameRenderer(Ogre::RenderWindow* renderWindow, Ogre::Root* root, Ogre::SceneTypeMask typeMask, const Ogre::String& instanceName) :
		GameRenderer(renderWindow, root, typeMask, instanceName),
		mViewport(NULL),
		mCameraMode(CM_THIRD_PERSON_CAMERA),
		mCamera1(NULL),
		mCamera3(NULL),
		mCameraObj(NULL),
		mCameraObjTargetBody(NULL),
		mDistanceCamera3(10),
		mTotalYawCamera1(0),
		mTotalPitchCamera1(0),
		mCameraObjTimer(0),
		mCameraObjLength(1),
		mCameraObjTranslation(Ogre::Vector3::ZERO),
		mCameraObjPreviousMode(CM_THIRD_PERSON_CAMERA),
		mCameraObjAnimationType(AT_FIXED),
		mChangeCameraTimer(0),
		mCameraChangePressed(false)
	{
		// Configure lights and shadows.
		//mSceneManager->setShadowTechnique(Ogre::SHADOWTYPE_STENCIL_MODULATIVE);
		mSceneManager->setAmbientLight(Ogre::ColourValue(0.3, 0.3, 0.25));
		Ogre::Light* light = mSceneManager->createLight("Sun");
		light->setType(Ogre::Light::LT_DIRECTIONAL);
		light->setDirection(-1, -1, -1);
		light->setDiffuseColour(1, 1, 0.90);
		light->setSpecularColour(1, 1, 0.95);

		// Third person camera (default camera).
		mCamera3 = mSceneManager->createCamera("3rdPersonRubyCamera");
		mViewport = renderWindow->addViewport(mCamera3);
		mViewport->setBackgroundColour(Ogre::ColourValue());
		mCamera3->setAspectRatio(Ogre::Real(mViewport->getActualWidth())/Ogre::Real(mViewport->getActualHeight()));
		mCamera3->setNearClipDistance(0.1);
		mCamera3->setFarClipDistance(3000);
		mCamera3->setPosition(0, 0, 0);

		// First person camera.
		mCamera1 = mSceneManager->createCamera("1stPersonRubyCamera");
		mCamera1->setAspectRatio(Ogre::Real(mViewport->getActualWidth())/Ogre::Real(mViewport->getActualHeight()));
		mCamera1->setNearClipDistance(0.1);
		mCamera1->setFarClipDistance(3000);
		mCamera1->setPosition(0, 0, 0);

		// Object camera.
		mCameraObj = mSceneManager->createCamera("ObjectRubyCamera");
		mCameraObj->setAspectRatio(Ogre::Real(mViewport->getActualWidth())/Ogre::Real(mViewport->getActualHeight()));
		mCameraObj->setNearClipDistance(0.1);
		mCameraObj->setFarClipDistance(3000);
		mCameraObj->setPosition(0, 0, 0);

		// Scene fog.
		Ogre::ColourValue fadeColour(1,1,1);
		mSceneManager->setFog(Ogre::FOG_LINEAR, fadeColour, 0.0, 300, 500 );
	    renderWindow->getViewport(0)->setBackgroundColour(fadeColour);

		// Reset overlays.
		Ogre::OverlayManager::getSingleton().getByName("Mask")->hide();
		Ogre::OverlayManager::getSingleton().getOverlayElement("MaskPanel")->hide();
	}

	RubyGameRenderer::~RubyGameRenderer()
	{

	}

	void RubyGameRenderer::update(double elapsedTime)
	{
		// Update the base.
		GameRenderer::update(elapsedTime);
		
		// Camera control.
		if(mController != NULL)
		{
			this->checkCameraChange(elapsedTime);

			// Thirst person camera management.
			if(mCameraMode == CM_THIRD_PERSON_CAMERA)
			{
				mCamera3->setPosition(mCharacterBody->getSceneNode()->getPosition());

				AbstractController* c2 = mController->getController2();
				mDistanceCamera3 += c2->getTranslationZ()*0.05;

				if(mDistanceCamera3 < Properties::CAMERA_3_MIN_DIST) 
					mDistanceCamera3 = Properties::CAMERA_3_MIN_DIST;

				Ogre::Quaternion r(Ogre::Degree(c2->getRotationAngle()),
								   Ogre::Vector3(c2->getRotationAxisX(), 
												 c2->getRotationAxisY(), 
												 c2->getRotationAxisZ()));
				mCamera3->yaw(r.getYaw()*0.1);
				mCamera3->pitch(r.getPitch()*0.1);
				mCamera3->moveRelative(Ogre::Vector3(0, 0, mDistanceCamera3));

				if(mCamera3->getPosition().y < Properties::CAMERA_3_MIN_ALTI)
				{
					mCamera3->setPosition(mCamera3->getPosition().x, Properties::CAMERA_3_MIN_ALTI, mCamera3->getPosition().z);
					mCamera3->lookAt(mCharacterBody->getSceneNode()->getPosition());
				}
			} // First person camera management.
			else if(mCameraMode == CM_FIRST_PERSON_CAMERA)
			{
				Ogre::AxisAlignedBox box = mCharacterBody->getCollisionAABB();

				// Position of the camera is the "nose" of our character.
				mCamera1->setPosition(box.getCenter() + box.getHalfSize() * Ogre::Vector3::NEGATIVE_UNIT_Z);

				AbstractController* c2 = mController->getController2();

				Ogre::Quaternion r(Ogre::Degree(c2->getRotationAngle()),
								   Ogre::Vector3(c2->getRotationAxisX(), 
												 c2->getRotationAxisY(), 
												 c2->getRotationAxisZ()));

				mTotalYawCamera1 += r.getYaw().valueDegrees() * 0.1;
				mTotalPitchCamera1 += r.getPitch().valueDegrees() * 0.1;
					
				if(mTotalYawCamera1 > Properties::CAMERA_1_MAX_YAW) mTotalYawCamera1 = Properties::CAMERA_1_MAX_YAW;	
				else if(mTotalYawCamera1 < -Properties::CAMERA_1_MAX_YAW) mTotalYawCamera1 = -Properties::CAMERA_1_MAX_YAW;
				if(mTotalPitchCamera1 > Properties::CAMERA_1_MAX_PITCH) mTotalPitchCamera1 = Properties::CAMERA_1_MAX_PITCH;
				else if(mTotalPitchCamera1 < -Properties::CAMERA_1_MAX_PITCH) mTotalPitchCamera1 = -Properties::CAMERA_1_MAX_PITCH;
				mCamera1->setOrientation(Ogre::Quaternion::IDENTITY);
				mCamera1->yaw(Ogre::Degree(mTotalYawCamera1));
				mCamera1->pitch(Ogre::Degree(mTotalPitchCamera1));

			} // Object camera management.
			else if(mCameraMode == CM_OBJECT_CAMERA && mCameraObjTargetBody != NULL)
			{
				mCameraObjTimer += elapsedTime;
				
				if(mCameraObjTimer < mCameraObjLength)
				{
					if(mCameraObjAnimationType == AT_AROUND)
					{
						double angle = 360. * mCameraObjTimer / (mCameraObjLength + 0.0001);
						mCameraObj->setOrientation(Ogre::Quaternion::IDENTITY);
						mCameraObj->setPosition(mCameraObjTargetBody->getSceneNode()->getPosition());
						mCameraObj->yaw(Ogre::Degree(angle));
						mCameraObj->moveRelative(mCameraObjTranslation);
						mCameraObj->lookAt(mCharacterBody->getSceneNode()->getPosition());
					}
					else if (mCameraObjAnimationType == AT_FIXED)
					{
						mCameraObj->setPosition(mCameraObjTargetBody->getSceneNode()->getPosition() + mCameraObjTranslation);
						mCameraObj->lookAt(mCharacterBody->getSceneNode()->getPosition());
					}
				}
				else
				{
					this->setCameraMode(mCameraObjPreviousMode);
				}
			}
		}
	}


	void RubyGameRenderer::setCameraMode(CameraMode cm)
	{ 
		if(cm == CM_THIRD_PERSON_CAMERA)
		{		
			Ogre::OverlayManager::getSingleton().getByName("Mask")->hide();
			Ogre::OverlayManager::getSingleton().getOverlayElement("MaskPanel")->hide();
			mViewport->setCamera(mCamera3);
			mCharacterBody->getSceneNode()->setVisible(true);
		} 
		else if(cm == CM_FIRST_PERSON_CAMERA)
		{
			// Body's invisible to prevent from seeing faces of the model.
			mViewport->setCamera(mCamera1);
			mCharacterBody->getSceneNode()->setVisible(false);

			// Ruby's mask
			Ogre::OverlayManager::getSingleton().getByName("Mask")->show();
			Ogre::OverlayManager::getSingleton().getOverlayElement("MaskPanel")->show();
		} 
		else if(cm == CM_OBJECT_CAMERA)
		{
			Ogre::OverlayManager::getSingleton().getByName("Mask")->hide();
			Ogre::OverlayManager::getSingleton().getOverlayElement("MaskPanel")->hide();
			mViewport->setCamera(mCameraObj);
			mCharacterBody->getSceneNode()->setVisible(true);
			mCameraObjTimer = 0;

			if(mCameraMode != CM_OBJECT_CAMERA)
				mCameraObjPreviousMode = mCameraMode;
		}
		
		mCameraMode = cm; 
	}

	void RubyGameRenderer::setCharacterBody(ObjectBody* characterBody)
	{ 
		GameRenderer::setCharacterBody(characterBody);
		mEarNode = characterBody->getSceneNode();
	}

	void RubyGameRenderer::setCameraObjectOptions(ObjectBody* targetBody, CameraObjectAnimationType animationType, double length, Ogre::Vector3& translation)
	{
		mCameraObjTargetBody = targetBody;
		mCameraObjLength = length;
		mCameraObjTranslation = translation;
		mCameraObjAnimationType = animationType;
	}

	void RubyGameRenderer::checkCameraChange(double elapsedTime)
	{
		AbstractController* c3 = mController->getController3();
		
		if(c3->getAction1() > 0 && mChangeCameraTimer == 0 && mCameraChangePressed == false)
		{
			if(mCameraMode == CM_THIRD_PERSON_CAMERA)
				this->setCameraMode(CM_FIRST_PERSON_CAMERA);
			else if(mCameraMode == CM_FIRST_PERSON_CAMERA)
				this->setCameraMode(CM_THIRD_PERSON_CAMERA);

			mCameraChangePressed = true;
		}

		if(mCameraChangePressed == true)
		{
			mChangeCameraTimer += elapsedTime;
			if(mChangeCameraTimer > 0.2)
			{
				mChangeCameraTimer = 0;
				mCameraChangePressed = false;
			}
		}
	}
}
